/* TrafficVisor HTTP daemon */

/* -------- Default Settings -------- */

AppNumber       = 0                                  /* see "-N" TV's option */
address.!port   = 80                                 /* listen port          */
address.!addr   = 'INADDR_ANY'                       /* listen address       */
TVConf          = 'tvisor.cfg'                       /* TV's config file     */
RefreshDataTime = 5                                  /* data refresh period
						        in seconds           */
/* After "RefreshDataTime" period statistics */
/* will be reread from TrafficVisor's pipe   */


say 'TrafficVisor II HTTP daemon'
say

/* -------- Read config flie -------- */

arg FCfg
FCfg  = space(FCfg)
if FCfg='-?' | FCfg='-h' | FCfg='-H' then
do
  say 'Usage:'
  say ' TVHTTPd [CfgFile]'
  say 'By default, CfgFile is "TVHTTPd.cfg".'
  exit
end
if FCfg = '' then FCfg = 'TVHTTPd.cfg'

if stream(FCfg,'c','query exists')='' then
do
  say 'File 'FCfg' not found.'
  exit
end

NLine = 0
do while lines(FCfg)
  NLine = NLine+1
  parse upper value linein(FCfg) with Key Val
  if length(Key)=0 | substr(Key,1,1)='#' then iterate
  Val = space(Val)
  select
    when Key='APPNUMBER' then 
    do
      if DataType(Val,N) then AppNumber = Val
      else CfgError('Invalid "AppNumber" value')
    end
    when Key='PORT' then 
    do
      if DataType(Val,N) then address.!port = Val
      else CfgError('Invalid port')
    end
    when Key='BINDADDR' then 
    do
      if Val = '*' then address.!addr = 'INADDR_ANY'
      else address.!addr = Val
    end
    when Key='REFRESHDATATIME' then 
    do
      if DataType(Val,N) then RefreshDataTime = Val
      else CfgError('Invalid "RefreshDataTime" value')
    end
    when Key='TVCONF' then TVConf = Val
  otherwise
    CfgError('Unknown key "'Key'"')
  end
end



call rxfuncadd sysloadfuncs, rexxutil, sysloadfuncs;
call sysloadfuncs;

if RxFuncQuery("SockLoadFuncs") then
do
  call RxFuncAdd "SockLoadFuncs","rxSock","SockLoadFuncs"
  call SockLoadFuncs
end

PipeName = '\PIPE\TrafficVisor'AppNumber'\MAIN'
HistoryCount = 0

StatFile=''
do while lines(TVConf)
  str = linein(TVConf)
  if translate(word(str,1))='STATFILE' then StatFile = word(str,words(str))
end
call stream TVConf,'c','close'
if StatFile='' then StatFile = 'statistics.out'
else StatFile = strip(StatFile,'B','"')

rc = SockInit()
numeric digits 20

/* ----- Initialization server socket ----- */

sock = SockSocket(AF_INET, SOCK_STREAM, IPPROTO_TCP)
opt = x2d('FFFFFFF')
rc = SockSetSockOpt(sock, 'SOL_SOCKET', 'SO_REUSEPORT', opt);
address.!family = 'AF_INET'
rc = SockBind(sock,"address.!")
if rc\=0 then
do
  SockPSock_Errno('bind()')
  exit
end
rc = SockListen(sock,5)

say 'Started'


/* /----- SERVER CYCLE -----\ */

rc = GetTVStatistics(PipeName)
rc = time('R')
do forever
  ClientSock = SockAccept(sock, "client.!");
  say 'Client connected: 'client.!addr

  drop str
  rc = SockRecv(ClientSock,str,2048,'msg_peek')
  if rc\=-1 then
  do
    str = substr(str,pos('/',str)+1)
    str = substr(str,1,pos(' ',str)-1)
    parse var str Doc '?' SortField

    rc = SendLine('<html><head></head><body>')
    rc = SendHistoryList()
    if Doc\='' then rc = GetTVStatistics(StatFile'.'Doc)
    else do
      if (time('E')%1)>=RefreshDataTime then
      do
        rc = GetTVStatistics(PipeName)
        rc = time('R')
      end
    end
    rc = SendInfo()
    rc = SendLine('</body></html>')

    rc = SockSoClose(ClientSock) 
    rc = SockShutdown(ClientSock,2)
  end
end

/* \----- SERVER CYCLE -----/ */

rc = SockShutdown(ClientSock, 2) 
rc = SockShutdown(sock, 2) 
EXIT



/* ----------------------- FUNCTIONS ----------------------- */

SendLine: procedure expose ClientSock
  LineCR = arg(1)''x2c('0D0A')
return SockSend(ClientSock,LineCR,0)

SendInfo: procedure expose ClientSock PipeName groups. SortField Period LastSortField
  NeedSort = (LastSortField\=SortField)
  rc = SendLine('<center>'Period'<table border=0 cellspacing=0 cellpadding=0>')
  if groups.0>0 then
  do i = 1 to groups.0
    rc = SendLine('<tr><td colspan=3 align=center><br>'groups.i.name'</td></tr>')
    str =    '<td align=center><font size=-1><a href="?name">Name</a></font></td>'
    str = str'<td align=center><font size=-1><a href="?send">Send</a></font></td>'
    str = str'<td align=center><font size=-1><a href="?recv">Receive</a></font></td>'
    rc = SendLine('<tr bgcolor=#D0EED0>'str'</tr>')
    if NeedSort then rc = Sort(i)
    do j = 1 to groups.i.list.0
      str =    '<td align=left><font size=-1>'groups.i.list.j.name'</font>&nbsp</td>'
      str = str'<td align=right><font size=-1>'BToStr(groups.i.list.j.1)'</font>&nbsp</td>'
      str = str'<td align=right><font size=-1>'BToStr(groups.i.list.j.0)'</font>&nbsp</td>'
      if j//2=0 then color = 'E8EBD4' 
      else color = 'F2F3E0'
      rc = SendLine('<tr bgcolor=#'color'>'str'</tr>')
    end
  end
  rc = SendLine('</table></center>')
  LastSortField = SortField
return 0

GetTVStatistics: procedure expose groups. Period PipeName LastSortField History. StatFile
  FName = arg(1)

  drop History.
  call SysFileTree StatFile'.???','History','FO'

  do i = 1 to History.0
    P = lastpos('.',History.i)
    History.i = substr(History.i,P+1)
  end
  do i = 1 to History.0
    do j = 1 to History.0-1
      k = j+1
      if History.j>History.k then
      do
        s = History.j
        History.j = History.k
        History.k = s
      end
    end
  end

  drop LastSortField
  drop groups.
  groups.0 = 0

  if FName=PipeName then rc = lineout(FName,'comments')
  if rc<0 then return rc
  i = 0
  do while lines(FName)
    i = i+1
    line.i = linein(FName)
  end
  line.0 = i
  rc = stream(FName,'c','close')
  if i<2 then return -1
  drop Period
  Period = line.1
  P = pos('Period:',Period)
  Period = substr(Period,P)

  InRule = 0
  do i = 2 to line.0
    str = line.i
    if InRule=0 then
    do
      InRule = (word(str,1)='rule')
      CommentValid = 0
    end
    else do
      if word(str,1)='end' then InRule = 0
      else do
        key = word(str,1)
        if key='Comment' then
        do 
          str = subword(str,2,100)
          P = pos(':',str)
          if P=0 then iterate 
          Group = substr(str,1,P-1)
          Send = substr(str,P+1,1)
          Send = (Send='s' | Send='S' | Send='1')
          P = pos(':',str,P+1)
          if P=0 then iterate
          Course = substr(str,P+1)
          Group = GetGroup(Group)
          Course = GetCourse(Group,Course)
          CommentValid = 1
        end
        else if (key='Bytes') & CommentValid then
        do
          groups.Group.list.Course.Send = word(str,2)
        end
      end
    end
  end
return 0

GetGroup: procedure expose groups.
  N = arg(1)
  do j = 1 to groups.0
    if groups.j.name = N then return j
  end
  j = groups.0+1
  groups.0 = j
  groups.j.name = N
  groups.j.list.0 = 0
return j

GetCourse: procedure expose groups.
  Index = arg(1)
  N = arg(2)
  do j = 1 to groups.Index.list.0
    if groups.Index.list.j.name = N then return j
  end
  j = groups.Index.list.0+1
  groups.Index.list.0 = j
  groups.Index.list.j.name = N
  groups.Index.list.j.0 = ''
  groups.Index.list.j.1 = ''
return j

Sort: procedure expose groups. SortField
  Index = arg(1)
  do i = 1 to groups.Index.list.0
    do j = 1 to groups.Index.list.0-1
      k = j+1
      nName = groups.Index.list.k.name
      nSend = groups.Index.list.k.1
      nRecv = groups.Index.list.k.0

      cName = groups.Index.list.j.name
      cSend = groups.Index.list.j.1
      cRecv = groups.Index.list.j.0

      select
        when SortField='name' then comp = cName>nName
        when SortField='send' then comp = cSend<nSend
        otherwise             comp = cRecv<nRecv
      end
      if comp then
      do
        groups.Index.list.k.name = cName
        groups.Index.list.k.1    = cSend
        groups.Index.list.k.0    = cRecv
        groups.Index.list.j.name = nName
        groups.Index.list.j.1    = nSend
        groups.Index.list.j.0    = nRecv
      end
    end
  end
return 0

SendHistoryList: procedure expose ClientSock History. StatFile Doc SortField
  if History.0=0 then return 0
  str = '<a href="/?'SortField'">[Now]</a>&nbsp'
  do i = 1 to History.0
    if History.i=Doc then str = str'<font color=red>['i']</font>&nbsp'
    else str = str'<a href="/'History.i'?'SortField'">['i']</a>&nbsp'
  end
  rc = SendLine('<center><font size=-1>'str'</font></center>')
return 0

BToStr:
  B = arg(1)
  if B>1024000000 then return trunc(B/1024000000,2)' Gb'
  if B>1024000 then return trunc(B/1024000,2)' Mb'
  if B>1024 then return trunc(B/1024,2)' Kb'
  return B' Bytes'
return 0

CfgError:
    say arg(1)' in file 'FCfg' at line 'NLine
  exit